
import unittest
import tempfile
import shutil
import itertools
from collections import defaultdict

import numpy as np

import pyemma.coordinates as coor
import pyemma.coordinates.tests.util as util
from pyemma.coordinates.data import FragmentedTrajectoryReader


def max_chunksize_from_config(itemsize):
    from pyemma import config
    from pyemma.util.units import string_to_bytes
    max_bytes = string_to_bytes(config.default_chunksize)
    max_frames = max(1, int(np.floor(max_bytes / itemsize)))
    return max_frames


class GenerateTestMatrix(type):
    def __new__(mcs, name, bases, attr):
        from functools import partial

        # needed for python2
        class partialmethod(partial):
            def __get__(self, instance, owner):
                if instance is None:
                    return self
                return partial(self.func, instance,
                               *(self.args or ()), **(self.keywords or {}))
        new_test_methods = {}

        test_templates = {k: v for k, v in attr.items() if k.startswith('_test') }
        test_parameters = attr['params']
        for test, params in test_templates.items():
            test_param = test_parameters[test]
            for param_set in test_param:
                func = partialmethod(attr[test], **param_set)
                # only 'primitive' types should be used as part of test name.
                vals_str = '_'.join((str(v) if not isinstance(v, np.ndarray) else 'array' for v in param_set.values()))
                assert '[' not in vals_str, 'this char makes pytest think it has to extract parameters out of the testname.'
                out_name = '{}_{}'.format(test[1:], vals_str)
                func.__qualname__ = 'TestReaders.{}'.format(out_name)
                new_test_methods[out_name] = func

        attr.update(new_test_methods)
        return type.__new__(mcs, name, bases, attr)


class TestReaders(unittest.TestCase, metaclass=GenerateTestMatrix):
    """
    trajectory lengths:
        - 5000
        - 6000
        - 1000

    to cover:
    - optimized lagged iterator
    - legacy lagged iterator
    - io-optimization in patches on and off

    - file formats:
        - csv (PyCSVReader)
        - npy (NumpyFileReader)
        - in-memory (DataInMemory)
        - mdtraj-supported formats (FeatureReader)

    - chunk sizes:
        - 0 (whole traj)
        - None (whole traj)
        - 1 (very small chunk size)
        - 100 (medium chunk size)
        - 10000000 (chunk size exceeding file frames)

    - lag:
        - no lag
        - 1 (very small lag)
        - 100 (moderate lag)
        - 5000 (lag so that it exceeds the file frames for at least a couple trajs)

    - fragmented reader
    - sources merger

    The tests:
        - The lag/no-lag cases can be covered in separate tests as we should test if the lagged data is correct
        - When iterating without lag we just have to check that each chunk is <= chunksize (or even == chunksize) and
          that the data matches
        - separate test cases for fragmented reader and sources merger
    """

    tempdir = None
    n_atoms = 6
    n_dims = 6 * 3

    chunk_sizes = (0, 1, 127, 5000, 10000, None)
    strides = (1, 3, 100)
    ra_strides = (
        np.array([[0, 1], [0, 3], [0, 3], [0, 5], [0, 6], [0, 7], [2, 1], [2, 1]]),
        np.array([[0, 23], [0, 42], [0, 4999], [1, 3], [2, 666], [3, 42], [4, 128]])
    )
    skips = (0, 123)
    lags = (1, 128, 5000)
    file_formats = ('in-memory',
                    'numpy',
                    'xtc',
                    'trr',
                    'dcd',
                    'h5',  # pyemma H5Reader
                    'csv',
                    )
    # transform data or not (identity does not change the data, but pushes it through a StreamingTransformer).
    transforms = (None, 'identity')

    eps = np.finfo(np.float32).eps

    # pytest config
    params = {
        '_test_base_reader': [dict(file_format=f, stride=s, skip=skip, chunksize=cs, transform=t)
                              for f, s, skip, cs,t  in itertools.product(file_formats, strides, skips,
                                                                         chunk_sizes, transforms)],
        '_test_lagged_reader': [
            dict(file_format=f, stride=s, skip=skip, chunksize=cs, lag=lag)
            for f, s, skip, cs, lag in itertools.product(file_formats, strides, skips, chunk_sizes, lags)
        ],
        '_test_fragment_reader': [
            dict(file_format=f, stride=s, lag=l, chunksize=cs)
            for f, s, l, cs in itertools.product(file_formats, strides, [0] + list(lags), chunk_sizes)
        ],
        '_test_base_reader_with_random_access_stride': [
          dict(file_format=f, stride=s, chunksize=cs)
            for f, s, cs in itertools.product(file_formats, ra_strides, chunk_sizes)
        ]
    }

    @classmethod
    def setup_class(cls):
        cls.file_creators = {
            'csv': util.create_trajectory_csv,
            'in-memory': lambda dirname, data: data,
            'numpy': util.create_trajectory_numpy,
            'xtc': lambda dirname, data: util.create_trajectory_xtc(cls.n_atoms, dirname, data),
            'trr': lambda dirname, data: util.create_trajectory_trr(cls.n_atoms, dirname, data),
            'dcd': lambda dirname, data: util.create_trajectory_dcd(cls.n_atoms, dirname, data),
            'h5': lambda dirname, data: util.create_trajectory_h5(cls.n_atoms, dirname, data)
        }
        cls.tempdir = tempfile.mkdtemp("test-api-src")
        cls.traj_data = [np.random.random((5000, cls.n_dims)).astype(np.float32),
                         np.random.random((50, cls.n_dims)).astype(np.float32),
                         np.random.random((6000, cls.n_dims)).astype(np.float32),
                         np.random.random((125, cls.n_dims)).astype(np.float32),
                         np.random.random((1000, cls.n_dims)).astype(np.float32)]
        # trajectory files for the different formats
        cls.test_trajs = {}
        for file_format in cls.file_formats:
            assert file_format in cls.file_creators.keys()
            cls.test_trajs[file_format] = [cls.file_creators[file_format](cls.tempdir, X) for X in cls.traj_data]

        cls.pdb_file = util.create_dummy_pdb(cls.tempdir, cls.n_dims // 3)

    @classmethod
    def teardown_class(cls):
        shutil.rmtree(cls.tempdir, ignore_errors=True)

    # H5Reader does not need a topology (and gets only selected in case we do not pass it).
    def setUp(self):
        if 'h5' in self._testMethodName:
            self.pdb_file_bak = self.pdb_file
            self.pdb_file = None

    def tearDown(self):
        if 'h5' in self._testMethodName:
            self.pdb_file = self.pdb_file_bak

    def _test_lagged_reader(self, file_format, stride, skip, chunksize, lag):
        trajs = self.test_trajs[file_format]
        reader = coor.source(trajs, top=self.pdb_file, chunksize=chunksize)

        it = reader.iterator(stride=stride, skip=skip, lag=lag, chunk=chunksize)
        traj_data = [data[skip::stride] for data in self.traj_data]
        traj_data_lagged = [data[skip + lag::stride] for data in self.traj_data]
        valid_itrajs = [i for i, x in enumerate(traj_data_lagged) if len(x) > 0]

        assert it.chunksize is not None
        if chunksize is None:
            chunksize = max_chunksize_from_config(reader.output_type().itemsize)

        with it:
            current_itraj = None
            t = t_total = 0
            collected = defaultdict(list)
            collected_lag = defaultdict(list)

            for itraj, chunk, chunk_lagged in it:
                # reset t upon next trajectory
                if itraj != current_itraj:
                    current_itraj = itraj
                    t = 0
                assert len(chunk) <= chunksize or chunksize == 0
                if chunksize != 0 and len(traj_data[itraj]) - t >= chunksize:
                    assert len(chunk) <= chunksize
                elif chunksize == 0:
                    assert len(chunk) == len(chunk_lagged) == len(traj_data_lagged[itraj])
                collected[itraj].append(chunk)
                collected_lag[itraj].append(chunk_lagged)

                t += len(chunk)
                t_total += len(chunk)

        for itraj in valid_itrajs:
            assert itraj in collected.keys()
            assert itraj in collected_lag.keys()

        assert set(collected.keys()) == set(collected_lag.keys())
        for itraj in collected.keys():
            assert itraj in valid_itrajs
            collected[itraj] = np.vstack(collected[itraj])
            collected_lag[itraj] = np.vstack(collected_lag[itraj])
            # unlagged data is truncated to the length of the lagged data.
            max_len = len(traj_data_lagged[itraj])
            np.testing.assert_allclose(collected[itraj], traj_data[itraj][:max_len], atol=self.eps)
            np.testing.assert_allclose(collected_lag[itraj], traj_data_lagged[itraj], atol=self.eps)

        assert t_total == sum(len(x) for x in collected.values())
        assert t_total == reader.n_frames_total(stride=stride, skip=skip+lag)

    def _test_fragment_reader(self, file_format, stride, lag, chunksize):
        trajs = self.test_trajs[file_format]

        reader = coor.source([trajs], top=self.pdb_file, chunksize=chunksize)
        assert isinstance(reader, FragmentedTrajectoryReader)

        data = np.vstack(self.traj_data)
        itraj = None

        if lag > 0:
            collected = []
            collected_lagged = []
            for itraj, X, Y in reader.iterator(stride=stride, lag=lag):
                collected.append(X)
                collected_lagged.append(Y)
            assert collected
            assert collected_lagged
            assert len(collected) == len(collected_lagged)
            collected = np.vstack(collected)
            collected_lagged = np.vstack(collected_lagged)
            np.testing.assert_allclose(data[::stride][0:len(collected_lagged)], collected, atol=self.eps,
                                                    err_msg="lag={}, stride={}, cs={}".format(lag, stride, chunksize
                                                 ))
            np.testing.assert_allclose(data[lag::stride], collected_lagged, atol=self.eps)
        else:
            collected = []
            for itraj, X in reader.iterator(stride=stride):
                collected.append(X)
            assert collected
            collected = np.vstack(collected)
            np.testing.assert_allclose(data[::stride], collected, atol=self.eps)
            assert itraj == 0 # only one trajectory

    def _test_base_reader(self, file_format, stride, skip, chunksize, transform):
        trajs = self.test_trajs[file_format]
        reader = coor.source(trajs, top=self.pdb_file, chunksize=chunksize)

        if transform == 'identity':
            reader = util.create_transform(reader)

        if chunksize is not None:
            np.testing.assert_equal(reader.chunksize, chunksize)

        it = reader.iterator(stride=stride, skip=skip, lag=0, chunk=chunksize)

        assert it.chunksize is not None
        if chunksize is None:
            max_frames = max_chunksize_from_config(reader.output_type().itemsize)
            assert it.chunksize <= max_frames
            # now we set the chunksize to max_frames, to be able to compare the actual shapes of iterator output.
            chunksize = max_frames

        traj_data = [data[skip::stride] for data in self.traj_data]
        valid_itraj = [i for i, x in enumerate(traj_data) if len(x) > 0]
        output = defaultdict(list)

        with it:
            current_itraj = None
            t = t_total = 0
            for itraj, chunk in it:
                # reset t upon next trajectory
                if itraj != current_itraj:
                    current_itraj = itraj
                    t = 0

                assert len(chunk) <= chunksize or chunksize == 0, '%s' % it
                if chunksize != 0 and len(traj_data[itraj]) - t >= chunksize:
                    assert len(chunk) == chunksize
                elif chunksize == 0:
                    assert len(chunk) == len(traj_data[itraj])

                output[itraj].append(chunk)

                t += len(chunk)
                t_total += len(chunk)

            for itraj in valid_itraj:
                assert itraj in output.keys()

            for itraj in output.keys():
                assert itraj in valid_itraj
                output[itraj] = np.vstack(output[itraj])
                np.testing.assert_allclose(output[itraj], traj_data[itraj], atol=self.eps)

            assert t_total == sum(len(x) for x in output.values())
            assert t_total == reader.n_frames_total(stride=stride, skip=skip)

    def _test_base_reader_with_random_access_stride(self, file_format, stride, chunksize):
        trajs = self.test_trajs[file_format]
        reader = coor.source(trajs, top=self.pdb_file, chunksize=chunksize)

        if chunksize is not None:
            np.testing.assert_equal(reader.chunksize, chunksize)

        it = reader.iterator(stride=stride, lag=0, chunk=chunksize)

        assert it.chunksize is not None
        if chunksize is None:
            max_frames = max_chunksize_from_config(reader.output_type().itemsize)
            assert it.chunksize <= max_frames
            # now we set the chunksize to max_frames, to be able to compare the actual shapes of iterator output.
            chunksize = max_frames
        traj_data = [data[stride[stride[:, 0] == i][:, 1]] for i, data in enumerate(self.traj_data)]

        with it:
            current_itraj = None
            t = t_total = 0
            for itraj, chunk in it:
                # reset t upon next trajectory
                if itraj != current_itraj:
                    current_itraj = itraj
                    t = 0

                assert len(chunk) <= chunksize or chunksize == 0
                if chunksize != 0 and len(traj_data[itraj]) - t >= chunksize:
                    assert len(chunk) == chunksize
                elif chunksize == 0:
                    assert len(chunk) == len(traj_data[itraj])

                np.testing.assert_allclose(chunk, traj_data[itraj][t:t+len(chunk)], atol=self.eps)

                t += len(chunk)
                t_total += len(chunk)
            assert t_total == reader.n_frames_total(stride=stride, skip=0) == it.n_frames_total()


if __name__ == '__main__':
    unittest.main()



# This file is part of PyEMMA.
#
# Copyright (c) 2017 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy as np

from pyemma.coordinates import tica, tica_nystroem
from pyemma.coordinates.transform.nystroem_tica import oASIS_Nystroem

import sys
use_assert_warns = (sys.version_info >= (3, 4))


class TestNystroemTICA_Simple(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.data = np.ones((10000, 100))
        cls.variable_columns = np.random.choice(100, 10, replace=False)
        cls.data[:, cls.variable_columns] = np.random.rand(10000, 10)
        # Start with one of the constant columns:
        cls.initial_columns = np.setdiff1d(np.arange(cls.data.shape[1]), cls.variable_columns)[0:1]
        cls.tica_obj = tica(data=cls.data, lag=1)

    def _compare_tica(self, t):
        np.testing.assert_allclose(t._oasis.error, 0, rtol=1e-05, atol=1e-08)
        np.testing.assert_allclose(t.cov, self.tica_obj.cov[:, t.column_indices])
        np.testing.assert_allclose(t.cov_tau, self.tica_obj.cov_tau[:, t.column_indices])
        np.testing.assert_allclose(t.timescales, self.tica_obj.timescales)

    def test_single(self):
        t = tica_nystroem(data=self.data, lag=1, max_columns=11, initial_columns=self.initial_columns)
        self._compare_tica(t)

    def test_single_issues_warning(self):
        if use_assert_warns:
            with self.assertLogs(level='WARNING'):
                t = tica_nystroem(data=self.data, lag=1, max_columns=12, initial_columns=self.initial_columns)
        else:
            t = tica_nystroem(data=self.data, lag=1, max_columns=12, initial_columns=self.initial_columns)
        self._compare_tica(t)

    def test_multiple(self):
        t = tica_nystroem(data=self.data, lag=1, max_columns=11, nsel=3,
                          initial_columns=self.initial_columns)
        self._compare_tica(t)

    def test_multiple_issues_warning(self):
        if use_assert_warns:
            with self.assertWarns(UserWarning):
                tica_nystroem(data=np.random.rand(100, 10), lag=1, max_columns=11, nsel=3,
                              initial_columns=np.random.choice(10, 2, replace=False))

    def test_describe(self):
        # just check there is no exception
        tica_nystroem(max_columns=2).describe()
        tica_nystroem(max_columns=2, data=np.random.random((100, 10)))


class TestNystroemTICA_DoubleWell(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        from pyemma.datasets import load_2well_discrete
        dw = load_2well_discrete()
        v = dw.dtraj_T100K_dt10[:10000]
        cls.T = v.size
        nstates = 100
        b = np.linspace(-1, 1, nstates)
        sigma = 0.15
        cls.Z = np.zeros((cls.T, nstates))
        for t in range(cls.T):
            for j in range(nstates):
                cls.Z[t, j] = np.exp(-(b[v[t]]-b[j])**2/(2*sigma**2))
        cls.lag = 10
        cls.tica_obj = tica(data=cls.Z, lag=cls.lag)

    def test(self):
        t = tica_nystroem(data=self.Z, lag=self.lag, max_columns=10, initial_columns=np.array([91]))
        np.testing.assert_allclose(t.cov, self.tica_obj.cov[:, t.column_indices])
        np.testing.assert_allclose(t.cov_tau, self.tica_obj.cov_tau[:, t.column_indices])
        assert np.all(np.abs(self.tica_obj.eigenvalues[:4] - t.eigenvalues[:4]) < 1e-3)


class TestNystroemTICA_oASIS(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.X = np.ones((10000, 10))
        cls.variable_columns = np.random.choice(10, 5, replace=False)
        cls.X[:, cls.variable_columns] = np.random.randn(10000, 5)
        cls.C0 = np.dot(cls.X.T, cls.X)
        cls.d = np.diag(cls.C0)
        cls.initial_columns = np.setdiff1d(np.arange(10), cls.variable_columns)[0:1]
        cls.C0_k = cls.C0[:, cls.initial_columns]

    def test_random(self):
        oasis = oASIS_Nystroem(self.d, self.C0_k, self.initial_columns)
        oasis.set_selection_strategy(strategy='random', nsel=1)
        while oasis.k < 6:
            newcol = oasis.select_columns()
            c = self.C0[:, newcol]
            added_columns = oasis.add_columns(c, newcol)
            assert ((len(added_columns) > 0 and np.any(np.in1d(self.variable_columns, added_columns)))
                   or (len(added_columns) == 0 and not np.all(np.in1d(self.variable_columns, newcol))))
        assert oasis.select_columns() is None
        np.testing.assert_allclose(oasis.error, 0, rtol=1e-08, atol=1e-10)

    def _test_oasis_single(self, strategy):
        oasis = oASIS_Nystroem(self.d, self.C0_k, self.initial_columns)
        oasis.set_selection_strategy(strategy=strategy, nsel=1)
        for _ in range(5):
            newcol = oasis.select_columns()
            c = self.C0[:, newcol]
            assert oasis.add_columns(c, newcol) == newcol
        assert oasis.select_columns() is None
        np.testing.assert_allclose(oasis.error, 0, rtol=1e-08, atol=1e-10)
        return oasis

    def test_oasis(self):
        oasis = self._test_oasis_single('oasis')
        spectral_oasis = self._test_oasis_single('spectral-oasis')
        assert np.all(np.equal(oasis.column_indices, spectral_oasis.column_indices))

    def test_spectral_oasis_multiple(self):
        oasis = oASIS_Nystroem(self.d, self.C0_k, self.initial_columns)
        oasis.set_selection_strategy(strategy='spectral-oasis', nsel=5)
        newcols = oasis.select_columns()
        c = self.C0[:, newcols]
        assert np.all(np.equal(oasis.add_columns(c, newcols), newcols))
        assert oasis.select_columns() is None
        np.testing.assert_allclose(oasis.error, 0, rtol=1e-08, atol=1e-10)

    def test_initial_zero(self):
        X = np.zeros((3, 3))
        X[:, 2:3] = np.ones((3, 1))
        C0 = np.dot(X.T, X)
        initial_columns = np.array([0, 1])
        oasis = oASIS_Nystroem(np.diag(C0), C0[:, initial_columns], initial_columns)
        np.testing.assert_allclose(np.abs(oasis.error), np.array([0, 0, 3]))
        np.testing.assert_allclose(oasis.approximate_matrix(), np.zeros((3, 3)))
        for strategy in ['random', 'oasis', 'spectral-oasis']:
            oasis.set_selection_strategy(strategy=strategy, nsel=1)
            assert oasis.select_columns() == np.array([2])

    def test_constant_evec(self):
        U = np.vstack([1+np.random.randn(1, 4)*1e-6,
                       np.array([1, -1, 0, 0]),
                       np.array([0, 0, 1, -1])]).T
        M = np.dot(U, np.dot(np.diag([1, 0.95, 0.7]), U.T))
        oasis = oASIS_Nystroem(np.diag(M), M[:, 0:2], np.array([0, 1]))
        from pyemma.coordinates.transform.nystroem_tica import SelectionStrategySpectralOasis
        sel = SelectionStrategySpectralOasis(oasis, strategy='spectral-oasis')
        evecs = oasis.approximate_eig()[1]
        fixed_evecs = sel._fix_constant_evec(evecs)
        assert np.all(fixed_evecs[:, 0] == 1)
        assert evecs.shape == fixed_evecs.shape


if __name__ == "__main__":
    unittest.main()



# This file is part of PyEMMA.
#
# Copyright (c) 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as _np


class SubSet(object):
    # note: we do not derive from SerializableMixIn, since this class is not intended to be saved (just its derivatives)
    __serialize_version = 0
    __serialize_fields = ('active_set', 'nstates_full')

    @property
    def active_set(self):
        r"""The active set of states on which all computations and estimations will be done."""
        if not hasattr(self, "_active_set"):
            raise AttributeError("active_set: the list of active states has not been given.")
        return self._active_set

    @active_set.setter
    def active_set(self, active_set):
        if active_set is None:
            active_set = []
        self._active_set = _np.asarray(active_set, dtype=_np.intc)

    @property
    def nstates_full(self):
        r"""Size of the full set of states."""
        if not hasattr(self, "_nstates_full"):
            raise AttributeError(
                "nstates_full: the size of the full set of states has not been given.")
        return self._nstates_full

    @nstates_full.setter
    def nstates_full(self, nstates_full):
        try:
            self._nstates_full = int(nstates_full)
        except TypeError:
            self._nstates_full = None


def _globalize(data, axis, active_set, default_value, n_centers):
    if data.ndim == 1:
        array = _np.asarray([default_value]).repeat(n_centers)
        array[active_set] = data
    elif data.ndim == 2:
        expanded_shape = list(data.shape)
        expanded_shape[axis] = n_centers
        array = _np.asarray([default_value]).repeat(
            expanded_shape[0] * expanded_shape[1]).reshape(expanded_shape)
        if axis == 0:
            array[active_set, :] = data
        elif axis == 1:
            array[:, active_set] = data
        else:
            raise NotImplementedError()
    else:
        raise NotImplementedError()
    return array


def _wrap_to_full_state(name, default_value, axis):
    from pyemma._base.estimator import _call_member

    def alias_to_full_state(self, *args, **kw):
        data = _call_member(self, name, *args, **kw)
        data = _np.asarray(data)
        try:
            return _globalize(data, axis, self.active_set, default_value, self.nstates_full)
        except AttributeError:
            return data

    return alias_to_full_state


def add_full_state_methods(class_with_globalize_methods):
    """
    class decorator to create "_full_state" methods/properties on the class (so they
    are valid for all instances created from this class).

    Parameters
    ----------
    class_with_globalize_methods

    """
    assert hasattr(class_with_globalize_methods, 'active_set')
    assert hasattr(class_with_globalize_methods, 'nstates_full')

    for name, method in class_with_globalize_methods.__dict__.copy().items():
        if isinstance(method, property) and hasattr(method.fget, '_map_to_full_state_def_arg'):
            default_value = method.fget._map_to_full_state_def_arg
            axis = method.fget._map_to_full_state_along_axis
            new_getter = _wrap_to_full_state(name, default_value, axis)
            alias_to_full_state_inst = property(new_getter)
        elif hasattr(method, '_map_to_full_state_def_arg'):
            default_value = method._map_to_full_state_def_arg
            axis = method._map_to_full_state_along_axis
            alias_to_full_state_inst = _wrap_to_full_state(name, default_value, axis)
        else:
            continue

        name += "_full_state"
        setattr(class_with_globalize_methods, name, alias_to_full_state_inst)

    return class_with_globalize_methods


class map_to_full_state(object):
    """ adds a copy of decorated method/property to be passed to the full state interpolation function

    Parameters
    ----------
    default_arg: object
        the default argument to interpolate missing values with.
    extend_along_axis : int, default=0
        extend along given axis for multi-dimensional data.
    """

    def __init__(self, default_arg, extend_along_axis=0):
        self.default_arg = default_arg
        self.extend_along_axis = extend_along_axis

    def __call__(self, func):
        if isinstance(func, property):
            func.fget._map_to_full_state_def_arg = self.default_arg
            func.fget._map_to_full_state_along_axis = self.extend_along_axis
        else:
            func._map_to_full_state_def_arg = self.default_arg
            func._map_to_full_state_along_axis = self.extend_along_axis
        return func




def get_n_jobs(logger=None):

    def _from_hardware():
        import psutil
        return psutil.cpu_count(logical=False)

    def _from_env(var):
        import os
        e = os.getenv(var, None)
        if e:
            try:
                return int(e)
            except ValueError as ve:
                if logger is not None:
                    logger.warning("could not parse env variable '{var}'."
                                   " Value='{val}'. Error={err}."
                                   .format(err=ve, val=e, var=var))
        return None

    slurm_njobs = _from_env('SLURM_CPUS_ON_NODE')  # Number of CPUS on the allocated SLURM node.
    pyemma_njobs = _from_env('PYEMMA_NJOBS')

    if slurm_njobs and pyemma_njobs:
        import warnings
        warnings.warn('two settings for n_jobs from environment: PYEMMA_NJOBS and SLURM_CPUS_ON_NODE. '
                      'Respecting the SLURM setting to avoid overprovisioning resources.')

    # slurm njobs will be used preferably.
    val = slurm_njobs or pyemma_njobs
    if not val:
        val = _from_hardware()
    if logger is not None:
        logger.debug('determined n_jobs: %s', val)
    return val


class NJobsMixIn(object):
    # mixin for sklearn-like estimators (estimation/ctor parameter has to contain n_jobs).

    @property
    def n_jobs(self):
        """ Returns number of jobs/threads to use during assignment of data.

        Returns
        -------
        If None it will return the setting of 'PYEMMA_NJOBS' or
        'SLURM_CPUS_ON_NODE' environment variable. If none of these environment variables exist,
        the number of processors /or cores is returned.

        Notes
        -----
        This setting will effectively be multiplied by the the number of threads used by NumPy for
        algorithms which use multiple processes. So take care if you choose this manually.
        """
        if not hasattr(self, '_n_jobs'):
            self._n_jobs = get_n_jobs(logger=getattr(self, 'logger'))
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, val):
        if val is None:
            val = get_n_jobs(logger=getattr(self, 'logger'))
        self._n_jobs = int(val)




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


r"""
===============================================================================
transform - Transformation Utilities (:mod:`pyemma.coordinates.transform`)
===============================================================================
.. currentmodule:: pyemma.coordinates.transform

.. autosummary::
    :toctree: generated/

    PCA - principal components
    TICA - time independent components
    VAMP - Variational approach for Markov processes
    VAMPModel - Kinetic model form the Variational approach for Markov processes
    VAMPChapmanKolmogorovValidator - Chapman Kolmogorov test for the Variational approach for Markov processes
"""

from .pca import *
from .tica import *
from .nystroem_tica import *
from .vamp import *
